/*
			Copyright (c) 1994 by
			Advanced Visual Systems Inc.
			All Rights Reserved
	
	This software comprises unpublished confidential information of
	Advanced Visual Systems Inc. and may not be used, copied or made
	available to anyone, except in accordance with the license
	under which it is furnished.
	
	This file is under Perforce control
	$Id: //depot/express/fcs992/modules/slice.c#1 $
*/
#include <avs/util.h>
#include <avs/err.h>
#include <avs/db.h>
#include <avs/fld.h>
#include <avs/math.h>

#define ERR_RETURN(A) ERRerror("slice",1,ERR_ORIG, A); return(0);
#define MAX_NAME_SIZE 1024
#define FIELDS_IN_SYNC(A,B) 0
	
static float IXFORM[16] = {1.0, 0.0, 0.0, 0.0,
			   0.0, 1.0, 0.0, 0.0,
			   0.0, 0.0, 1.0, 0.0,
			   0.0, 0.0, 0.0, 1.0};
slice(elem_id)
    OMobj_id elem_id;
{
	OMobj_id in, out, plane, param;

	if (OMis_null_obj(elem_id)) {
		fprintf (stderr, "\n** SLICE: module is 0**\n");
		return(0);
	}

	in = OMfind_subobj(elem_id, OMstr_to_name("in"), OM_OBJ_RD);
	if (OMis_null_obj(in)) {
		fprintf (stderr, "\n** SLICE: in is 0**\n");
		return(0);
	}
	out = OMfind_subobj(elem_id, OMstr_to_name("out"), OM_OBJ_RW);
	if (OMis_null_obj(out)) {
		fprintf (stderr, "\n** SLICE: out is 0**\n");
		return(0);
	}
	plane = OMfind_subobj(elem_id, OMstr_to_name("plane"), OM_OBJ_RD);
	if (OMis_null_obj(plane)) {
		fprintf (stderr, "\n** SLICE: plane is 0**\n");
		return(0);
	}
	param = OMfind_subobj(elem_id, OMstr_to_name("nparam"), OM_OBJ_RW);
	if (OMis_null_obj(param)) {
		fprintf (stderr, "\n** SLICE: plane is 0**\n");
		return(0);
	}
	if (!FUNCslice(in, plane, param, out)) {
		fprintf (stderr, "\n**slice returns 0**\n");
	}

	return(1);
}

int FUNCslice (in, plane, param, out)
    OMobj_id  in, plane, out, param;
{
	int   nnodes, out_ncells, out_nnodes, type, size, conn_size, veclen, ncomp;
	int   *node_list, *out_nlist, *min_node_list, *max_node_list;
	int   npts, stat, i;
	char  units[MAX_NAME_SIZE];
	int   seq_in, seq_out;
	float  *coord, *out_coord, *t, *pnts;
	float  bounds[4][3], norm[3], norm_xfm[3], xfm[4][4], *xform;
	float  *node_dist, plane_dist;
	char   *cell_table;
	OMobj_id   cell_set;

	seq_in = OMget_obj_seq(in, OMnull_obj, 0);
	seq_out = OMget_obj_seq(out, OMnull_obj, 0);

	out_nnodes = 0;
	/*-----------------*/
	/*   GET PLANE     */
	/*-----------------*/	
	if (FLDget_points (plane, &pnts, &size, OM_GET_ARRAY_RD) != 1) {
		ERR_RETURN("Error getting points");
	}
	if (UTILget_plane_unif(pnts, bounds, norm) != 1) {
		ERR_RETURN("Error calculating slice plane");
	}
	stat = FLDget_xform(plane, &xform);
	if (stat < 0) {
		ERR_RETURN("cannot get xform for probe");
	}
	else if (stat == 0)
		xform = IXFORM;

	/*------------------*/
	/* TRANSFORM PLANE  */
	/*------------------*/
	for (i=0; i<4; i++) 
		memcpy(xfm[i], xform+4*i, 4*sizeof(float));
	for (i=0; i<4; i++) 
		MATvec3_mat4_multiply(bounds[i], xfm);
	MATxform_vecs(1, norm, xfm, norm_xfm);
	plane_dist = VEC_DOT(bounds[0], norm_xfm);

	if (1) {
		if (FLDget_nnodes(in, &nnodes) != 1) {
			ERR_RETURN("cannot get nnodes");
		}
		if (FLDget_coord(in, &coord, &size, OM_GET_ARRAY_RD) != 1) {
			ERR_RETURN("cannot get coordinates");
		}
		
		node_dist = (float *)malloc(nnodes * sizeof(float));
		for (i=0; i<nnodes; i++)
			node_dist[i] = VEC_DOT((coord+3*i), norm_xfm);
		
		stat = FUTILcell_table_create(in, &cell_table);
		if (!stat) {
			ERR_RETURN("cannot create cell table");
		}

		UTILslice_cells(nnodes, cell_table, node_dist, OM_TYPE_FLOAT,
				(double)plane_dist,
				&out_nnodes, &min_node_list, &max_node_list,
				&out_ncells, &out_nlist, &conn_size);

		FUTILcell_table_delete (cell_table);

		/*** OUTPUT FIELD ***/
		
		if (out_nnodes) {
			if (FLDget_coord_units(in, units, MAX_NAME_SIZE) != 1) {
				ERR_RETURN("Error getting units");
			}
			if (FLDset_coord_units (out, units) != 1) {
				ERR_RETURN("Error setting nspace");
			}
			if (FLDset_nspace (out, 3) != 1) {
				ERR_RETURN("Error setting nspace");
			}
			if (FLDset_nnodes (out, out_nnodes) != 1) {
				ERR_RETURN("Error setting nnodes");
			}
			
			if (FLDset_ncell_sets (out, 1) != 1) {
				ERR_RETURN("Error setting ncell_sets");
			}
			
			if (FLDget_cell_set(out, 0, &cell_set) != 1) {
				ERR_RETURN("Error getting cell set");
			}
			
			if (FLDset_cell_set(cell_set, "Tri") != 1) {
				ERR_RETURN("Error setting cell type");
			}
			
			if (FLDset_ncells(cell_set, out_ncells) != 1) {
				ERR_RETURN("Error setting ncells");
			}
			
			if (FLDget_coord(out, &out_coord, &size, OM_GET_ARRAY_WR) != 1) {
				ERR_RETURN("Error setting coordinate array");
			}
			
			if (FLDset_node_connect(cell_set, out_nlist, conn_size,
						    OM_SET_ARRAY_FREE) != 1) {
				ERR_RETURN("Error setting cell connect list");
			}
			
			if (FLDset_int (param, "nnodes", out_nnodes) != 1) { 
				ERR_RETURN("Error setting out nnodes");
			}
			if (FLDset_array_int (param, "min_node", out_nnodes, 
					       min_node_list, OM_SET_ARRAY_FREE) != 1) {
				ERR_RETURN("Error setting min_node_list");
			}
			if (FLDset_array_int (param, "max_node", out_nnodes, 
					       max_node_list, OM_SET_ARRAY_FREE) != 1) {
				ERR_RETURN("Error setting max_node_list");
			}
			if (FLDget_array_float (param, "tpar", &t, &size, 
						 OM_GET_ARRAY_WR) != 1) {
				ERR_RETURN("Error setting t array");
			}
			UTILget_param(out_nnodes, OM_TYPE_FLOAT, node_dist,
				      plane_dist, min_node_list, max_node_list, t); 
			UTILeval_param(out_nnodes, 3, OM_TYPE_FLOAT, coord,
				       min_node_list, max_node_list, t, out_coord);
		}
	}
	return(1);
}
